from abc import ABCMeta
from typing import Tuple, Optional, List, Dict, Iterable
from functools import partial

import cv2
import sys
import numpy as np
import logging

from dataspree.inference.ai.models.base_model import BaseModel, InferenceMode
from dataspree.inference.core.item import Item

from dataspree.platform_sdk.decoder.opencv_decoder import OpencvDecoder

logger = logging.getLogger(__name__)

try:
    from albumentations import ReplayCompose
    from dataspree.inference.processing.augmentations.transform import (OneOfRetainProb, RelativePadding, RelativeCrop,
                                                                        RelativeCenterCrop, ChannelAssembler)

except Exception as e:
    # Custom transformations could not be included (potentially because optional pacakges are not installed)
    # This is not a problem if there are no preprocessing augmentations to be performed.
    logger.info(f'The preprocessing module could not be imported ({e}). Thus, it is impossible to apply the same '
                'preprocessing steps as performed during training. If this feature is requested later on, '
                'Dataspree Inference will report an ImportError.')


class ImageBaseModel(BaseModel, metaclass=ABCMeta):

    def __init__(self, input_shape: Tuple[int, int, int, int] = (0, 0, 0, 0),
                 inference_mode: InferenceMode = InferenceMode.WHOLE_IMAGE,
                 selected_rois: Optional[List[str]] = None, preprocessing: Optional[Dict] = None):
        self.input_shape: Tuple[int, int, int, int] = input_shape
        self.inference_mode: InferenceMode = inference_mode
        self.selected_rois: List[str] = [] if selected_rois is None else selected_rois
        self.current_roi_order: List[str] = []
        self.preprocessing: Optional[Dict] = preprocessing

        if self.preprocessing and sys.modules.get('dataspree.inference.processing') is None:
            raise ImportError('This model is configured to apply preprocessing augmentations.'
                              'However, the required preprocessing package could not be loaded.')

    def preprocess_item(self, item: Item) -> Item:
        """
        Preprocess entire item independ

        :param item:
        :return:
        """
        # If the item already contains a preprocessed image, drop it.
        # Thus, if found in the item, the preprocessed image always contains the preprocessed image from the
        # most recently executed model in the pipeline or None, if no custom preprocessing is executed.
        if 'preprocessed_image' in item:
            del item['preprocessed_image']

        image = item.get('image')
        image_structured = item.get('image_structured')

        if image_structured is None and image is not None:
            image_structured = OpencvDecoder.create_sturctured_array(image)

        if image is None and (image_structured is None or not self.preprocessing):
            return item

        if self.preprocessing:

            # Remove the Resize operation in case of RoI inference
            if self.inference_mode in [InferenceMode.SELECTED_ROIS, InferenceMode.ALL_ROIS]:
                for i in reversed(range(len(self.preprocessing['transforms']))):
                    if self.preprocessing['transforms'][i]['__class_fullname__'] == 'Resize':
                        self.preprocessing['transforms'].pop(i)
            try:
                item['preprocessed_image'] = \
                    ReplayCompose.replay(self.preprocessing, image=image_structured, keypoints=[], instances=[],
                                         positions=[])['image']
            except KeyError as e:
                logger.warning(f'Could not replay training augmentations: {e}. '
                               f'The version of the preprocessing library that was used to augment the '
                               f'image might posess an augmentation ({e}), that was not implemented '
                               f'in the currently installed version of the preprocessing software. '
                               f'Please update the preprocessing software if that is the case.')
                raise e from None

        return item

    def preprocess(self, item: Item) -> Optional[np.ndarray]:

        image: Optional[np.ndarray] = item.get('preprocessed_image', item.get('image'))
        if image is None:
            logger.debug('Discarding item that does not possess an image. The key "image" has to be set. '
                         'If preprocessing is active, make sure that the key "image_structured" is set, because '
                         'this is the key from which we extract the item.')
            return None

        item['inference_base_image'] = image
        height, width, channels = image.shape

        image_patches: List[np.ndarray] = []

        if self.inference_mode == InferenceMode.WHOLE_IMAGE:
            image_patches.append(image)

        elif self.inference_mode == InferenceMode.ALL_ROIS or self.inference_mode == InferenceMode.SELECTED_ROIS:
            item_rois: Dict = item.get('rois', {})
            selected_rois: Iterable[str] = item_rois.keys() if self.inference_mode == InferenceMode.ALL_ROIS \
                else self.selected_rois
            self.current_roi_order = []

            for roi_name in selected_rois:
                roi = item_rois.get(roi_name)
                if roi is None:
                    continue
                x_left, x_right, y_top, y_bottom = roi
                if 0 <= x_left < x_right <= width and 0 <= y_top < y_bottom <= height:
                    image_patch = image[y_top:y_bottom, x_left:x_right, :]
                    image_patches.append(image_patch)
                    self.current_roi_order.append(roi_name)

        if len(image_patches) == 0:
            return None

        images = np.empty((len(image_patches), self.input_shape[1], self.input_shape[2], self.input_shape[3]),
                          dtype=np.float32)
        for i, image_patch in enumerate(image_patches):

            # convert to float
            image_patch = image_patch.astype(np.float32, copy=False)

            # resize and handling RGB/BGR(A) to grayscale and grayscale to multichannel
            image_patch = cv2.resize(image_patch, (self.input_shape[3], self.input_shape[2]))
            if len(image_patch.shape) < 3:
                image_patch = image_patch[..., True]
            if self.input_shape[1] == 1 and self.input_shape[1] != image_patch.shape[2]:
                image_patch = image_patch.mean(2)[..., True]
            elif self.input_shape[1] > 1 and image_patch.shape[2] <= 1:
                image_patch = np.dstack([image_patch] * self.input_shape[1])

            # transpose from HWC to CHW
            image_patch = image_patch.transpose((2, 0, 1))

            # normalize: image = (2.0 / 255.0) * image - 1
            images[i] = (2.0 / 255.0) * image_patch - 1

        return images
