import os
import re
import logging

from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union, Set, no_type_check

from typing_extensions import TypeAlias

from dataspree.inference.core.exceptions import ConfigError, ImplementationError

logger = logging.getLogger(__name__)


class ConfigFieldTypes(str, Enum):
    BOOL_INPUT = 'bool_input'
    DATE_INPUT = 'date_input'
    DOUBLE_INPUT = 'double_input'
    GROUP = 'group'
    GROUP_SELECTION = 'group_selection'
    GROUP_SWITCHABLE = 'group_switchable'
    INTEGER_INPUT = 'integer_input'
    JSON_TEXT_INPUT = 'json_text_input'
    LIST_INPUT = 'list_input'
    STRING_INPUT = 'string_input'
    PASSWORD_INPUT = 'password_input'
    FILE_INPUT = 'string_input'
    SELECTION_DROPDOWN = 'selection_dropdown'
    TEXT_AREA_INPUT = 'text_area_input'

    def __json__(self):
        return self.value


def get_expected_type(cfg: ConfigFieldTypes):
    if cfg == ConfigFieldTypes.BOOL_INPUT:
        return bool

    if cfg == ConfigFieldTypes.DOUBLE_INPUT:
        return float

    if cfg == ConfigFieldTypes.INTEGER_INPUT:
        return int

    if cfg == ConfigFieldTypes.STRING_INPUT or cfg == ConfigFieldTypes.PASSWORD_INPUT or \
            cfg == ConfigFieldTypes.JSON_TEXT_INPUT:
        return str

    if cfg == ConfigFieldTypes.LIST_INPUT:
        return list

    return object

def check_type(value, type_hint) -> Optional[bool]:
    """
    Check typing annotation for simple types.
    """

    is_container: bool = hasattr(type_hint, '__args__')

    template_types: Optional[Tuple[TypeAlias, ...]]
    if is_container:
        template_types = type_hint.__args__ if type_hint.__args__ is not None else tuple()
    else:
        template_types = (type_hint,)

    if template_types is None:
        return None

    if type(value) == type_hint or type_hint == TypeVar or type_hint == Any:
        return True

    if len(template_types) == 1 and type_hint == List[template_types] and type(value) is list:  # type: ignore
        return all(check_type(a, template_types[0]) for a in value)

    if len(template_types) == 2:
        if type_hint == Dict[template_types[0], template_types[1]]:  # type: ignore
            if type(value) is dict:
                return all(check_type(a, template_types[0]) for a in value.keys()) and \
                       all(check_type(a, template_types[1]) for a in value.values())

    if len(template_types) >= 1 and type_hint == Tuple[template_types]:
        return all(check_type(a, template_types[i]) for i, a in enumerate(value))

    if len(template_types) > 1 and type_hint == Union[template_types]:
        return any(check_type(value, t) for t in template_types)

    try:
        return len(template_types) <= 1 and isinstance(value, type_hint)
    except TypeError:
        return False


def check_file(file_semantics, path: str, required: bool = True) -> None:
    if not path:
        if required:
            raise ConfigError(f'{file_semantics} is required.')
    elif not os.path.exists(path):
        raise ConfigError(f'{file_semantics} {path} does not exist.')


MAX_QUEUE_SIZE_CONFIG_FIELD = {
    'name': 'max_queue_size',
    'type': ConfigFieldTypes.INTEGER_INPUT,
    'display_name': 'max. queue size',
    'default': 1,
    'description': 'Maximum size of the input queue of the node.'
}

RESIZE_CONFIG_FIELD = {
    'name': 'resize',
    'type': ConfigFieldTypes.GROUP_SWITCHABLE,
    'display_name': 'resize',
    'description': 'Resize images.',
    'fields': [
        {
            'name': 'width',
            'type': ConfigFieldTypes.INTEGER_INPUT,
            'display_name': 'width'
        },
        {
            'name': 'height',
            'type': ConfigFieldTypes.INTEGER_INPUT,
            'display_name': 'height'
        }
    ],
}


def parse_resize_field(resize: Optional[Union[Dict[str, float], List[float]]]) -> Optional[Tuple[int, int]]:
    if resize is None:
        return None

    width: float
    height: float
    if type(resize) == list:
        if len(resize) != 2:
            raise ConfigError('resize must contain two entries: "width" and "height".')
        width, height = resize

    elif type(resize) == dict:
        if resize.get('width') is None or resize.get('height') is None:
            raise ConfigError('resize must contain the entries "width" and "height"')
        width, height = resize['width'], resize['height']
    else:
        raise ConfigError('Unsupported type for parameter "resize" (only "list" and "dict" '
                          'are supported).')

    if width < 1.0 or height < 1.0:
        raise ConfigError('resize width and height must be larger than 1.0')

    return int(width), int(height)


BATCH_AGGREGATION_ATTRIBUTES_FIELD = \
    {
        'name': 'attributes',
        'display_name': 'Attributes',
        'type': ConfigFieldTypes.LIST_INPUT,
        'description': 'Add the path of all attributes that are to be aggregated and stored '
                       'with the batch output.',
        'fields': [
            {
                'name': 'attribute',
                'display_name': 'Complete Attribute Path (optional)',
                'type': ConfigFieldTypes.LIST_INPUT,
                'fields': [
                    {
                        'name': 'attribute',
                        'display_name': 'Attribute Name (optional)',
                        'type': ConfigFieldTypes.STRING_INPUT
                    }
                ]
            },
        ]
    }


def camel_to_snake(name):
    """
    Transform snake case
    """
    return re.sub(r'(?<!^)(?=[A-Z])', '_', name).lower()


__subclasses_tv = TypeVar('__subclasses_tv')


def subclasses(super_class: Type[__subclasses_tv]) -> Set[Type[__subclasses_tv]]:
    """
    Iteratively gather all descendents (allowing for multiple levels of inheritance) of the
    provided #super_class.
    """

    if not hasattr(super_class, '__subclasses__'):
        return set()

    list_subclasses: List[Type[__subclasses_tv]] = super_class.__subclasses__()
    checked_types: Set[Type] = set()
    i: int = 0
    while i < len(list_subclasses):
        curr = list_subclasses[i]
        if curr not in checked_types:
            checked_types.add(curr)
            list_subclasses.extend(curr.__subclasses__())
        i += 1

    return set(list_subclasses)


__get_implementation_type = TypeVar('__get_implementation_type')


def get_implementation(super_class: Type[__get_implementation_type], sub_class_name: Optional[str],
                       default: Optional[Type[__get_implementation_type]] = None, ignore_list=None,
                       continue_multiple: bool = False) -> Type[__get_implementation_type]:
    candidates: List[Type[__get_implementation_type]] = get_implementations(super_class=super_class,
                                                                            sub_class_name=sub_class_name,
                                                                            ignore_list=ignore_list)
    if len(candidates) > 1:
        msg = f'Name ambiguity for {super_class}\'s subclasses: ' \
              f'Multiple candidates {candidates} for {sub_class_name}.'
        if not continue_multiple:
            raise ImplementationError(msg)
        logger.warning(msg)

    elif len(candidates) == 0:
        if default is not None:
            return default

        if hasattr(super_class, 'default'):
            try:
                default = getattr(super_class, 'default')()
                if default is not None and issubclass(default, super_class):
                    return default
            except (TypeError, ValueError):
                pass

        raise ConfigError(f'Class for {sub_class_name} not found.')

    return candidates[0]


def join(ls, separator=', ', separator_last=' and '):
    if len(ls) <= 1:
        return separator.join(ls)
    return separator_last.join((separator.join(ls[:-1]), separator[-1]))


def get_implementations(super_class: Type[__get_implementation_type], sub_class_name: Optional[str] = None,
                        ignore_list=None) -> List[Type[__get_implementation_type]]:
    """
    Return list of sub classes of #super_class that match the #sub_class_name.
    If there is no naming ambiguity (*) , the returned list is either empty if the sub class is not
    found, or contains exactly one entry.

    (*) Naming ambiguities can arise if the developers discard the naming convention, and provide
    two sub classes of the same super_class, one of which following a camel case and the other
    one a snake case naming convention.
    """
    if ignore_list is None:
        ignore_list = list()

    return [a for a in subclasses(super_class) if
            a not in ignore_list and (sub_class_name is None or camel_to_snake(a.__name__) == sub_class_name)]


def pretty_class_name(type_handle: type) -> str:
    return getattr(type_handle, '__name__', str(type_handle))


def is_user_defined_class(cls):
    return hasattr(cls, '__class__') and ('__dict__' in dir(cls) or hasattr(cls, '__slots__')
                                          or issubclass(cls, Enum))


IMAGE_FILE_FORMAT = {
    'name': 'file_format',
    'type': ConfigFieldTypes.SELECTION_DROPDOWN,
    'display_name': 'file format',
    'options': [
        {'text': 'png', 'value': '.png'},
        {'text': 'jpg', 'value': '.jpg'},
        {'text': 'tif', 'value': '.tif'}
    ],
    'description': 'File format for images.'
}


def parse_image_file_format(config, key='file_format'):
    file_format = config.get(key, '.png')
    allowed_file_formats = ['.png', '.jpg', '.jpeg', '.tif']
    if file_format not in allowed_file_formats:
        raise ConfigError(
            f'The file format must be one of the following: {allowed_file_formats}')
    return file_format
