import inspect
import logging
import copy
import traceback
from enum import Enum

from abc import ABC, ABCMeta
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union, Type
from functools import reduce

from dataspree.inference.core.config import camel_to_snake, join, ConfigFieldTypes
from dataspree.inference.core.design_patterns import StackContext
from dataspree.inference.core.config.auto import instantiate_from_config
from dataspree.inference.core.decorators import Execution, fallback_implementation, hook
from dataspree.inference.core.exceptions import ConfigError, ImplementationError, ConfigurationNotRetrievable
from dataspree.inference.core.item import Item
from dataspree.inference.core.config.form_generator import ConfigOption, parse_config_field

logger = logging.getLogger(__name__)

operation_parse: str = 'parsers'
operation_set: str = 'setters'
operation_update: str = 'updaters'
operation_serialize: str = 'serializers'

operations: Tuple[str, str, str, str, str] = (
    operation_parse, operation_set, operation_update, operation_serialize, 'hooks')


class Recovery(int, Enum):
    NONE = 0
    FROM_CONFIG = 1
    LIVE = 2


# XXX: Check that updater don't have parameters etc.
# XXX: make sure that update loops are avoided.
# XXX: This implies that updaters are not allowed to set the value they are updating.

def register_operations(operations_to_register, entity, name):
    for operation in operations_to_register:
        setattr(entity, operation, copy.copy(getattr(entity, operation)))

    # Register operations
    for var_name in dir(entity):
        for operation in operations_to_register:
            func = getattr(entity, var_name)

            if not callable(func) or not hasattr(func, operation) \
                    or type(getattr(func, operation)) not in (set, frozenset):
                continue

            if not len(getattr(func, operation)):
                raise ConfigError(f'Invalid {operation} detected. {func} is not associated to any variable. '
                                  f'Preventing instantiation of ill-defined class {name}.')

            # XXX: this makes problems for windows.
            # XXX: Test this for embedded python under windows,
            # pip install dist/dataspree-inference.whl -t [embedded python] and execute. Error reported for VideoCaputre
            is_func_static: bool = isinstance(inspect.getattr_static(entity, var_name), staticmethod) \
                                   or getattr(func, '__ds_in_static_hack__', None) is not None

            if is_func_static != isinstance(entity, type):
                if operation != 'hooks':
                    raise ConfigError(f'Invalid {operation} detected. {func} has to be a '
                                      f'{"non-" if is_func_static else ""}static member function. '
                                      f'Preventing instantiation of ill-defined class {name}.')
                else:
                    continue

            for var in getattr(func, operation):
                if operation == 'updaters':
                    getattr(entity, operation).setdefault(var, set()).add(func)
                else:
                    if var in getattr(entity, operation) and getattr(entity, operation)[var] != func:
                        raise ConfigError(f'Ambiguous {operation} {func.__name__} '
                                          f'and {getattr(entity, operation)[var].__name__} detected for variable '
                                          f'{var}. Preventing instantiation of ill-defined class {name}. ')
                    getattr(entity, operation)[var] = func

            break


class ConfigurableMeta(ABCMeta):

    def __new__(mcs, name: str, bases: Tuple[type, ...], attrs: Dict[str, Any], **kwargs: Any):  # -> type: cython
        """
        :param name:  Name of the target class
        :param bases: Bases of the target class
        :param attrs:
        :param kwargs:
        """

        # Add reconfiguration attributes to the target class
        if attrs is None:
            attrs = dict()

        if type(attrs) != dict:
            raise ConfigError('Classes that are created from `Configurable` must provide attrs as dict.'
                              f'Preventing instantiation of ill-defined class {name}.')

        if any(operation in attrs for operation in operations) or 'hooks' in attrs:
            raise ConfigError(f'Classes that are created from `Configurable` must not define {join(operations)} '
                              f'as their attributes. Preventing instantiation of ill-defined class {name}.')

        attrs.update({op: dict() for op in operations})
        for base in bases:
            attrs.update({op: dict(getattr(base, op, dict())) for op in operations})

        # Add logger to the target class
        if 'logger' in attrs:
            raise ConfigError(f'Classes that are created from `Configurable` must not override attribute `logger`. '
                              f'Preventing instantiation of ill-defined class {name}.')

        attrs['logger'] = logging.getLogger(name)

        if 'initialized' in attrs:
            raise ConfigError(f'Classes that are created from `Configurable` must not override attribute `initialized`.'
                              f' Preventing instantiation of ill-defined class {name}.')
        attrs['initialized'] = False

        if 'configuration_initialized' in attrs:
            raise ConfigError(f'Classes that are created from `Configurable` must not override attribute '
                              f'`configuration_initialized`.'
                              f' Preventing instantiation of ill-defined class {name}.')
        # attrs['configuration_initialized'] = dict()

        for a_k, a_v in attrs.items():
            try:
                if (a_v.__isabstractmethod__ or getattr(a_v, '__default_abstract__', False)) and hasattr(mcs, a_k):

                    super_func = getattr(mcs, a_k)
                    if super_func.__fallback_implementation__:
                        super_sig = set(list(inspect.signature(super_func).parameters)[1:])
                        a_sig = set(inspect.signature(a_v.__func__).parameters)
                        if len(super_sig) == len(a_sig) and all((a in super_sig for a in a_sig)):
                            attrs[a_k] = classmethod(super_func)

            except Exception as e:
                pass

        #
        # Create the class
        # Cython complains if type hitned.
        if '__getattribute_implementation__' in attrs and '__getattribute__' not in attrs:
            attrs['__getattribute__'] = attrs['__getattribute_implementation__']  # <  Pycharm linting fix
        cls = super(ConfigurableMeta, mcs).__new__(mcs, name, bases, attrs, **kwargs)
        register_operations([operation_parse, 'hooks'], cls, cls.__name__)

        #
        # Log tracing information.
        registered_operations_string: List[str] = []
        for op in operations:

            variables = getattr(cls, op)
            number_of_variables = len(variables)
            if number_of_variables:
                op_string = op[: - 1 if number_of_variables == 1 else None]
                attribute = 'the attribute' if number_of_variables == 1 else 'attributes'
                attributes = ', '.join(map(lambda a: '/'.join(a) if type(a) is not str else a, variables.keys()))
                registered_operations_string.append(f'{op_string} for {attribute} {attributes}')

        if len(registered_operations_string):
            attrs['logger'].debug(f'Registered {" and ".join(registered_operations_string)} '
                                  f'for configurable class {name}.')

        return cls

    def __call__(cls, *args, **kwargs):
        """
        Instantiates (evokes __init__) the class associated with the current meta class.

        Assembles configuration decorators in a way that is .

        :param args:   Unnamed constructor arguments of the class associated with the current meta class.
        :param kwargs: Named constructor arguments of the class associated with the current meta class.
        :return:
        """
        instance = super().__call__(*args, **kwargs)
        instance.initialized = True
        try:

            del instance.configuration_initialized
        except AttributeError as e:
            raise ImplementationError(f'Did not evoke the super constructor of Configurable: {e}')

        return instance

    def __getattribute__(cls, item):
        hooks: Optional[Union[Set, Dict]]
        try:
            hooks = super().__getattribute__('hooks')
        except AttributeError:
            hooks = None

        if hooks is not None and item in hooks:
            return cls.hooks[item](super().__getattribute__(item))

        return super().__getattribute__(item)

    def default_build(cls, config: Optional[Dict], from_user_input: bool = False, name: Optional[str] = None,
                      return_only_params: Optional[Set[str]] = None) \
            -> Optional['Configurable']:

        if config is None:
            return None

        pass_name = name is not None or 'name' in config

        if name is None:
            name = config.get('name')
        all_environment_variables = dict(name=name, from_user_input=from_user_input)

        item_config: Item = Item(config)

        # This evokes the build functionality
        # Evoke all the parsers
        for parser in getattr(cls, 'parsers').values():
            config_key: Union[Tuple, List]
            config_keys: List[Union[Tuple, List]] = parser.sources
            environment_variables: List[Any] = [all_environment_variables[evn] for evn in parser.env]
            config_values: List[Any] = [Item.get(config, config_key, ) for config_key in config_keys]

            item_config.update(parser(*config_values, *environment_variables))

        return instantiate_from_config(item_config, class_handle=cls,
                                       conditionals=dict(from_user_input=from_user_input, name=name),
                                       return_only_params=return_only_params)

    @fallback_implementation
    def build(cls, config: Optional[Dict], from_user_input: bool = False, name: Optional[str] = None) \
            -> Optional['Configurable']:

        return cls.default_build(config=config, from_user_input=from_user_input, name=name)


class Configurable(ABC, metaclass=ConfigurableMeta):

    def __init__(self):
        ABC.__init__(self)

        self.hooks: Dict[str, Callable] = getattr(self, 'hooks', dict())
        self.parsers: Dict[str, Callable] = getattr(self, 'parsers', dict())
        self.setters: Dict[str, Callable] = getattr(self, 'setters', dict())
        self.updaters: Dict[str, Callable] = getattr(self, 'updaters', dict())
        self.serializers: Dict[str, Callable] = getattr(self, 'serializers', dict())
        self.__got_configuration: bool = False

        self.last_configuration_options: Optional[Dict] = None

        self.__set_set: Set[str] = set()
        self.configuration_initialized: Dict[str, bool] = dict()
        register_operations([operation_set, operation_update, operation_serialize, 'hooks'], self, type(self).__name__)

    @classmethod
    def get_config_key(cls) -> str:
        return camel_to_snake(cls.__name__)

    @classmethod
    def get_config_entries(cls) -> Dict[str, Any]:
        """Return the properties with which this instances of this Node can be configured."""
        return dict()

    def get_configuration(self, report_error: bool = False) -> Dict:
        """
        Return the entries of the configuration of the instantiation of an entity (e.g. a Node).
        Replace the default values with the actual values.

        :raise ConfigurationNotRetrievable: if the configuration option could not be retrieved.
        """
        # Return the configuration entries and defaults for this node class.
        config_entries: Dict = copy.deepcopy(self.get_config_entries())

        # Emplace the config entries from last configuration
        config_entries_object: ConfigOption = parse_config_field(config_entries)
        config_entries_object.apply(self.last_configuration_options)
        gen_config_entries = Item(config_entries_object.generate_config_entries())
        for branch, new_value in Item(gen_config_entries).walk_leaves_and_branches(
                leaf_filter=lambda a, _: (len(a) and (a[-1] == 'default' or a[-1] == 'selection')),
                branch_filter=lambda a, _: (len(a) and (a[-1] == 'default' or a[-1] == 'selection'))):
            old_value = Item.get(config_entries, branch)
            ending = branch[-1]
            if new_value is not None:
                Item.set(config_entries, branch, new_value)
                Item.set(config_entries, branch[:-1] + ['default'], new_value)
                Item.set(config_entries, branch[:-1] + [f'{ending}_og'], old_value)
                Item.set(config_entries, branch[:-1] + ['recovered'], Recovery.FROM_CONFIG.value)

        # Update the class-defaults by fetching properties from the node instances.
        # Means of fetching class attributes are encoded in the `config_entries`.
        self.__emplace_config_of_field(self, config_entries, report_error=report_error)
        self.__got_configuration = True
        return config_entries

    def update_configuration(self, name: str, new_config: Dict, old_config_while_editing: Optional[Dict],
                             from_user_input: bool) -> bool:
        """
        :param name:            Name of the current entity
        :param new_config:      The new configuration
        :param old_config_while_editing:      The old configuration that was sent to the user and serves as a lookup to discern
                                changed values from unchanged values allowing the user to simultaneously edit
                                distinct properties.
        :return:
        """
        all_environment_variables = dict(name=name, from_user_input=from_user_input)

        if old_config_while_editing is None:
            return False

        old_config_while_editing: Item = Item(old_config_while_editing)
        new_config: Item = Item(new_config)

        # Determine the values that the user changed.
        changed_paths: Set[Tuple] = set()
        __remaining_paths: List[Tuple] = [tuple()]
        while len(__remaining_paths):
            abs_path = __remaining_paths.pop()
            new_value = new_config.get(list(abs_path))
            old_value = old_config_while_editing.get(list(abs_path))

            if isinstance(new_value, dict) and isinstance(old_value,
                                                          dict):  # XXX: this might be content of a configuration option!
                if len(new_value) and len(old_value):
                    __remaining_paths.extend([(*abs_path, k) for k in
                                              set(list(new_value.keys()) + list(old_value.keys()))])
                    continue

            if old_value != new_value:
                changed_paths.add(abs_path)

        del __remaining_paths

        # Assemble all instance variable names that are relevant for conducting the update
        own_attributes: Set[Tuple] = set()
        foreign_attributes: Set[Tuple] = set()
        parsers: Set[Callable] = set()

        own_values: Dict[Union[str, Tuple], Any] = \
            {next(iter(iv)) if len(iv) == 1 else iv: None for iv in own_attributes}

        for changed_path in changed_paths:
            try:
                if len(changed_path) and changed_path[-1] == 'selection':
                    changed_path = changed_path[:-1]

                if not len(changed_path):
                    raise ConfigurationNotRetrievable('Selection has to be reinstantiated')

                own, foreign, parser = self.get_instance_variable_names(changed_path)

                for f in own:
                    assert len(f) >= 1
                    # key = next(iter(f))
                    keys = f
                    default = object()
                    el = Item.get(new_config, list(f), default)

                    # if el is default:
                    #    raise ImplementationError(f'Key {key} missing in new configuration!')
                    if el is default:
                        own_values[keys[0] if len(keys) == 1 else keys] = None
                    else:
                        own_values[keys[0] if len(keys) == 1 else keys] = el

                foreign_attributes |= foreign
                own_attributes |= own
                if parser is not None:
                    parsers.add(parser)

            except ConfigurationNotRetrievable:
                return False
                # if changed_path[-1] != 'selection':
                #    # Not reconfigurable live
                #    return False
                # else:
                #    foreign_attributes |= {changed_path[:-1]} #XXX:

        # Determine if we can update the configuration inplace
        current_config_entries: Dict = self.get_config_entries()
        requires_accessors: bool = self.requires_accessors(current_config_entries)
        if requires_accessors and any((not self.has_accessor(o[0]) for o in own_attributes)):
            return False

        for parser in parsers:
            vals_parser = [Item.get(new_config, path) for path in parser.sources]
            environment_variables = [all_environment_variables[env] for env in parser.env]
            parser_dict = parser(*vals_parser, *environment_variables)

            # Prune away configurations in the dictionary that are extensions of a path reached via the parser.
            for k, v in parser_dict.items():
                if isinstance(k, tuple):
                    for k_c, v_c in own_values.items():
                        if isinstance(k_c, tuple) and len(k_c) > len(k) and k_c[:len(k)] == k:
                            own_values.pop(k_c, None)

                    if len(k) == 1:
                        own_values[k[0]] = v
                    else:
                        own_values[k] = v

                else:
                    own_values[k] = v

            # own_values.update({k[0] if isinstance(k, tuple) and len(k) == 1 else k: v for k, v in parser_dict.items()})

        values_setter: Dict[str, Any] = dict()
        values_manually_set: Dict[Tuple, Any] = dict()
        for k, v in own_values.items():
            (values_setter if type(k) == str else values_manually_set)[k] = v

        # Set attributes that are directly accessible and hence relevant for the setters
        self.set_attributes(**values_setter)

        # Set nested attributes that are irrelevant ofr the setters
        for ks, v in values_manually_set.items():
            Item.uni_setdefault(reduce(Item.uni_get, ks[:-1], self), ks[-1], v)

        # Forward values to child Configurables
        for ks in foreign_attributes:
            el = reduce(getattr, ks, self)

            foreign_node_name = getattr(el, 'name', None)
            foreign_new_config = Item.get(new_config, list(ks))
            foreign_old_config = old_config_while_editing.get(list(ks))

            if not el.update_configuration(name=foreign_node_name,
                                           new_config=foreign_new_config,
                                           old_config_while_editing=foreign_old_config,
                                           from_user_input=from_user_input):

                try:
                    # The following (type(el)) is wrong here as it may be too specialized.
                    # We would need the "official" type.
                    # sv = type(el).build(name=node_name, config=new_config, from_user_input=from_user_input)
                    # setattr(reduce(getattr, ks[:-1], self), ks[-1], sv)

                    # To obtain the "official" type we use the default build function to instantiate only
                    # the path to the subsequent object.
                    # XXX: Here I am not certain if the level of the configurable is okay.
                    sv = type(self).default_build(name=name, config=new_config, from_user_input=from_user_input,
                                                  return_only_params=[ks[0]])
                    setattr(self, ks[0], sv[ks[0]])
                    self.last_configuration_options = new_config
                    return True

                except Exception:
                    return False

        self.last_configuration_options = new_config
        return True

    def requires_accessors(self, config_entries: Optional[Dict] = None) -> bool:
        """
        Returns whether this class requires custom setters or updaters in order to change an attribute without
        needing to restart the plugin.
        """
        return (self.get_config_entries() if config_entries is None else config_entries).get('updatable') is not True

    def has_accessor(self, instance_variable_name: str):  # XXX: for other kind of updaters extend
        return instance_variable_name in self.setters or instance_variable_name in self.updaters

    def get_instance_variable_names(self, config_path: Tuple) -> Tuple[Set[Tuple], Set[Tuple], Optional[Callable]]:

        own_vars: Set[Tuple] = set()
        foreign_vars: Set[Tuple] = set()

        found_parser: Optional[Callable] = None
        changed_variables: Set[Tuple] = set()

        for parser in self.parsers.values():
            sources: List[List] = parser.sources

            source_path: List
            for source_path in sources:
                if len(source_path) == len(config_path) and \
                        all(config_path[i] == sp for i, sp in enumerate(source_path)):
                    if found_parser is parser:
                        continue

                    if found_parser is not None:
                        # XXX: is this ensured in the constructor?
                        raise ImplementationError(f'Parser configuration of class {type(self).__name__} is invalid.')

                    found_parser = parser

                    targets: Set[str] = parser.parsers
                    changed_variables |= {(t,) for t in targets}

        # Determine the associated configuration variable
        if found_parser is None:
            if self.has_accessor(config_path[-1]):
                changed_variables.add((config_path[-1],))
            elif self.has_accessor('_'.join((c for c in config_path if c is not None))):
                changed_variables.add(('_'.join((c for c in config_path if c is not None)),))
            else:

                el: Any = self
                pp: List[str] = list()
                for p in config_path:
                    pp.append(p)
                    el = getattr(el, p, None)

                    if isinstance(el, Configurable):
                        break

                if el is not None:
                    changed_variables.add(tuple(pp))
                else:
                    key, value = Configurable.__get_config_value(self, list(config_path), None, None, None)
                    (foreign_vars if isinstance(value, Configurable) else own_vars).add((key,) if isinstance(key, str) else tuple(key))

        for changed_variable_path in changed_variables:
            var = reduce(getattr, changed_variable_path, self)
            (foreign_vars if isinstance(var, Configurable) else own_vars).add(changed_variable_path)

        return own_vars, foreign_vars, found_parser

    def set_attributes(self, from_user_input: bool = True, **kwargs: Any) -> None:
        # Assemble all active setters that are required and directly set attributes without custom setter.
        active_setters: Set[Callable] = set()
        active_updaters: Set[Callable] = set()
        for k, v in kwargs.items():
            if k in self.setters and (self.initialized or all((self.configuration_initialized.get(var)
                                                               for var in self.setters[k].setters))):
                active_setters.add(self.setters[k])
            else:
                super(Configurable, self).__setattr__(k, v)

            if k in self.updaters and (self.initialized or all((self.configuration_initialized.get(var)
                                                                for var in self.updaters[k].updaters))):
                active_updaters |= self.updaters[k]

        # Evoke setters and updaters
        self.__evoke_setters(active_setters, from_user_input=from_user_input, **kwargs)
        self.__evoke_updaters(active_updaters)

    @staticmethod
    @hook('build')
    def inter_build(function, *args, **kwargs):
        config = copy.deepcopy(kwargs.get('config', args[0] if len(args) else None))
        instance = function(*args, **kwargs)
        if instance is not None:
            instance.last_configuration_options = copy.deepcopy(config)
        return instance

    def __getattribute_implementation__(self, item):
        retrieved_attribute = super().__getattribute__(item)

        hooks: Optional[Union[Set, Dict]]
        try:
            hooks = super().__getattribute__('hooks')
        except AttributeError:
            hooks = None

        if hooks is not None and item in hooks:
            return self.hooks[item](retrieved_attribute)

        return retrieved_attribute

    def __setattr__(self, key: str, value: Any) -> None:
        """
        :param key:    The key of the property to be set.
        :param value:  The new value the property is to assume.
        """
        with StackContext(getattr(self, '__set_set', None), key, f'Setting value in updater of variable {key}'):

            old_value = getattr(self, key, None)

            self.__setattr_set(key, value)
            self.__setattr_update(key)

            if key == 'initialized':
                definition: bool = old_value is None and not value
                initialization_terminated: bool = not old_value and value

                if not (definition or initialization_terminated):
                    raise ImplementationError(f'Variable {key} is reserved and cannot be set.')

                if initialization_terminated:
                    invalid_setters: Dict[str, List[str]] = dict()
                    remaining_setters: Set[callable] = set()
                    for setter in self.setters.values():
                        for var in setter.setters:
                            if not self.initialized and not self.configuration_initialized.get(var):
                                invalid_setters.setdefault(setter.__name__, list()).append(var)
                        if setter.execution_in_init == Execution.late_execution:
                            remaining_setters.add(setter)

                    invalid_updaters: Dict[str, List[str]] = dict()
                    remaining_updaters: Set[callable] = set()
                    for var_updater in self.updaters.values():
                        for updater in var_updater:
                            for var in updater.updaters:
                                if not self.initialized and not self.configuration_initialized.get(var):
                                    invalid_updaters.setdefault(updater.__name__, list()).append(var)
                            if updater.execution_in_init == Execution.late_execution:
                                remaining_updaters.add(updater)

                    invalid_message: str = ''
                    if len(invalid_setters):
                        invalid_message += 'Variables registered for setters left uninitialized: ' \
                                           f'{"; ".join([k + ": " + ", ".join(v) for k, v in invalid_setters.items()])}.'
                    if len(invalid_updaters):
                        invalid_message += 'Variables registered for updaters left uninitialized: ' \
                                           f'{"; ".join([k + ": " + ", ".join(v) for k, v in invalid_updaters.items()])}.'

                    if len(invalid_message):
                        raise ImplementationError(f'{invalid_message} Preventing instantiation of ill-defined class '
                                                  f'{type(self).__name__} {self.name}.')

                    self.__evoke_setters(remaining_setters)
                    self.__evoke_updaters(remaining_updaters)

    def __setattr_set(self, key: str, value: Any) -> None:
        """
        :param key:    The key of the property to be set.
        :param value:  The new value the property is to assume.
        """

        if key not in self.setters:
            return super(Configurable, self).__setattr__(key, value)

        # Custom setter available
        setter = self.setters[key]

        if id(self) in setter.active:
            return super(Configurable, self).__setattr__(key, value)

        # Custom setter available and not currently executed

        if self.initialized:
            raise ImplementationError(f'The attribute {key} possesses a custom setter. Direct assignment is only '
                                      f'permitted inside the constructor to avoid .')

        # Custom setter available, not currently executed and currently initializing

        super(Configurable, self).__setattr__(key, value)
        self.configuration_initialized[key] = True

        # Custom setter available, not currently executed, currently initializing and custom setter is to be evoked
        # if possible.
        if setter.execution_in_init == Execution.early_execution:
            if self.initialized or all((self.configuration_initialized.get(var) for var in setter)):
                self.__evoke_setters((setter,), **{key: value})

    def __setattr_update(self, key) -> None:
        """
        :param key:    The key of the property to be set.
        """

        for updater in self.updaters.get(key, set()):
            if not id(self) in getattr(self.setters.get(key), 'active', set()):

                if not self.initialized:
                    self.configuration_initialized[key] = True

                if self.initialized or updater.execution_in_init == Execution.early_execution:
                    if self.initialized or all((self.configuration_initialized.get(var) for var in updater.updaters)):
                        self.__evoke_updaters(self.updaters[key])

    def __evoke_updaters(self, updaters: Set[Callable[[], None]]):
        for u in updaters:
            u()

    def __evoke_setters(self, setters: Union[Set, Tuple], from_user_input: bool = False, **kwargs):

        # Fetch all the variables that need to be acquired for evoking the setter
        required_variables: List[str] = list()

        for setter in setters:
            required_variables.extend((variable for variable in setter.setters if variable not in kwargs.keys()))

        instance_variable_assignments: Dict[str, Any] = dict()

        # Apply the new values
        instance_variable_assignments.update(kwargs)

        # Apply the values that are easy to get
        instance_variable_assignments.update({v: getattr(self, v) for v in required_variables
                                              if v not in instance_variable_assignments})

        # Evoke all setters.
        for setter in setters:
            setter(**{name: assignment for name, assignment in instance_variable_assignments.items()
                      if name in setter.setters})

    def __emplace_config_of_field(self, entity: 'Configurable', field: Dict, path: Optional[List[str]] = None,
                                  skip_name: bool = False, value: Optional[Any] = None,
                                  raise_exception: bool = False, report_error: bool = False) -> bool:
        """
        Insert the node's current configuration of the `field`'s root as 'default' value.
        Also insert a flag 'recovered' that indicates that recovery was successful.

        :param field: Dictionary, at the level of the field that is to be recovered in this
                      iteration.

        :param path:  Path of the name of the variable that is to be recovered.
                      The configuration options of a node are nested, as configuration options
                      can be part of `groups`, `group_switchable`s or other configuration groupings.
                      path contains the absolute path of all the names from the configuration's
                      root to the current element that is to be recovered.

        :param skip_name: For specific properties the name of the current entity is not part of the
                          structure in `entity`. If that is the case, set skip_name to true.

        :raise ConfigurationNotRetrievable: if the configuration option could not be retrieved.
        """

        if value is not None:
            if isinstance(value, Configurable):
                path = None
                entity = value
                value = None

        if path is None:
            path = list()

        # Name of the entry that is to be recovered
        name: Optional[str] = None if skip_name else field.get('name')

        # Path to the entry that is to be recovered
        path_child: List[str] = path + [name] if name else path

        # Variable's data type that is to be recovered.
        current_type: Optional[ConfigFieldTypes] = field.get('type')

        # Human readable representation of the current path in the configuration dictionary used
        # to print coherent error messages.
        human_readable_path: str = f'in [{",".join(path)}]' if len(path) else ''
        human_readable_id: str = 'root' if current_type is None else f'{current_type.name} {name} {human_readable_path}'

        def handle_error(message: str) -> bool:
            if raise_exception:
                raise ConfigurationNotRetrievable(f'{human_readable_id}: {message}')

            if report_error and not self.__got_configuration:
                logger.warning(f'{human_readable_id}: {message}')

            return False

        # User-defined custom getter implementation
        set_custom_get: Set[Callable[[], Any]] = set(
            filter(lambda a: a is not None,
                   {field.get('getter'), self.__get_accessor('serializers', entity, tuple(path_child))}))
        if len(set_custom_get) > 1:
            return handle_error(f'Ambiguous call to getter. Candidates: {set_custom_get}. Aborting.')
        custom_get: Optional[Callable] = set_custom_get.pop() if len(set_custom_get) else None

        # Utilize reverse mapping from parser to sources for attempting to serialize
        parsed_variable_names: Optional[List[str]] = None
        if custom_get is None:
            parsed_variable_names = [variable_name for variable_name, parser in entity.parsers.items()
                                     if path_child in parser.sources]

        recovered: int = False
        recovered_value: Optional[Any] = None
        try:
            _recovered_key, recovered_value = self.__get_config_value(entity, path_child, name, custom_get, value=value)
            recovered = True
        except ConfigurationNotRetrievable as e:
            if custom_get is not None:
                return handle_error(f'The custom getter for field {name} reported a configuration error: {e}.')

            if len(parsed_variable_names) == 1:
                try:
                    _recovered_key, recovered_value = self.__get_config_value(entity, parsed_variable_names,
                                                                              None, None, None)
                    recovered = Recovery.LIVE.value
                except ConfigurationNotRetrievable:
                    ...

        if current_type in {ConfigFieldTypes.GROUP_SWITCHABLE, ConfigFieldTypes.GROUP,
                            ConfigFieldTypes.GROUP_SELECTION, None}:

            if current_type == ConfigFieldTypes.GROUP_SELECTION:
                # GROUP_SELECTION: Find out which option is selected Valid iff exactly one option
                #                  can be recovered.
                #                  Custom selection getter have the highest priority
                selection_name: Optional[str]
                try:
                    selection_name = recovered_value.get_config_key()
                except Exception as e:
                    selection_name = None

                kids = field.get('groups', [])
                if not len(kids):
                    return handle_error(f'{human_readable_id} defined without options.')

                successful = set()
                for i, f in enumerate(kids):
                    try:
                        if selection_name is None or f.get('name') == selection_name:
                            self.__emplace_config_of_field(entity, f, path_child, skip_name=False,
                                                           value=recovered_value,
                                                           raise_exception=True, report_error=report_error)
                            successful.add(i)

                            if selection_name is not None:
                                break

                    except ConfigurationNotRetrievable:
                        pass

                if len(successful) != 1:
                    filtered_success = [s for s in successful if kids[s].get('name') == field.get('default')
                                        and kids[s].get('name') is not None]

                    if len(filtered_success) == 1:
                        successful = set(filtered_success)

                    if len(successful) == 0:
                        return handle_error(f'No field could be recovered.')
                    if len(successful) > 1:
                        return handle_error(f'Multiple fields recovered simultaneously.')

                field['default'] = kids[successful.pop()].get('name')
                field['recovered'] = Recovery.LIVE.value
            else:
                # GROUP_SWITCHABLE: switched on iff all of the children can be recovered.
                # GROUP:            Recovered   iff all of the children can be recovered.

                fail = False
                for f in field.get('fields', []):
                    try:
                        self.__emplace_config_of_field(entity, f, path_child, value=recovered_value,
                                                       raise_exception=True, report_error=report_error)
                    except ConfigurationNotRetrievable as e:
                        fail = True
                        if current_type != ConfigFieldTypes.GROUP_SWITCHABLE:
                            handle_error(f'{e}')
                if fail:
                    field['default'] = False
                    field['recovered'] = Recovery.LIVE.value
                else:
                    if current_type == ConfigFieldTypes.GROUP_SWITCHABLE:
                        field['default'] = True
                    field['recovered'] = Recovery.LIVE.value

        else:
            # "Ordinary" configuration fields that can be retrieved directly.
            try:
                if not recovered or recovered_value is None:
                    if field.get('recovered') == Recovery.FROM_CONFIG:
                        return True

                    raise ConfigurationNotRetrievable(f'Could not find configuration of field {name}.')

                if isinstance(recovered_value, Enum):
                    field['default'] = recovered_value.value
                else:
                    field['default'] = recovered_value

                field['recovered'] = Recovery.LIVE.value
            except ConfigurationNotRetrievable as e:
                return handle_error(f'{e}')

        return True

    @staticmethod
    def __get_accessor(accessor_id, entity, var_name) -> Optional[Callable]:
        return None if entity is None else getattr(entity, accessor_id, dict()).get(var_name)

    @staticmethod
    def __get_config_value(entity, configuration_option: List[str],
                           new_configuration_option: Optional[str],
                           custom_get: Optional[Callable], value: Optional[Any]) -> Any:

        """
        Try to retrieve a specific configuration option from an existing and instantiated node.

        :param configuration_option: list of strings which identifies the configuration option, as
                                     extracted from the get_config_entries function.

        :param custom_get:   User-defined getter function that can be used to retrieve the config
                             option's assignment.

        :return:             The content that we would like to extract from an existing node
                             instance.

        :raise               ConfigurationNotRetrievable if the configuration option could not
                             be retrieved.
        """

        if not configuration_option:
            raise ConfigurationNotRetrievable('Empty path')

        entities = [entity]

        if value is not None:
            entities = [value, entity]

        if custom_get is not None:
            result_custom_get: Item = Item(custom_get())

            nc = Item.find(result_custom_get, [tuple(configuration_option)])
            if not nc[0]:
                nc = Item.find(result_custom_get, [(new_configuration_option,)])

            if not nc[0]:
                raise ConfigurationNotRetrievable(
                    f'Getter {getattr(custom_get, "__name__", "")} does not yield a value '
                    f'for {new_configuration_option}. '
                    f'Returned keys: {", ".join(".".join(f) for f in result_custom_get.keys())} ')
            return result_custom_get, nc[1]

        # Try to fetch the
        for ent in entities:
            try:
                return reduce(Configurable.__find_element, configuration_option, ent)
            except ConfigurationNotRetrievable:
                ...

        configuration_option_string = '_'.join([c for c in configuration_option])
        for ent in entities:
            try:
                return Configurable.__find_element(ent, configuration_option_string)
            except ConfigurationNotRetrievable:
                ...

        for ent in entities:
            try:
                # If the element has not been found, check if we can immediately access the most specialized path
                return Configurable.__find_element(ent, configuration_option[-1])
            except ConfigurationNotRetrievable:
                ...

        raise ConfigurationNotRetrievable(f'Could not retrieve a value for {configuration_option}.')

    @staticmethod
    def __find_element(current_root: Any, identifier_searched_item: str) -> Any:
        """
        Find and return the `identifier_searched_item` in the `current_root` element.
        This is used as an utility method for extracting an existing node's configuration.

        :param current_root: The current object which is searched for identifier_searched_item.
                             Might be a node object, or whatever we find in a node that is
                             stored under the specified field.

        :param identifier_searched_item:   The identifier of the element that we would like to
                                           extract.

        :return:             The content that we would like to extract from an existing node
                             instance.

        :raise               ConfigurationNotRetrievable if the configuration option could not
                             be retrieved.
        """
        if isinstance(current_root, dict) and identifier_searched_item in current_root:
            return identifier_searched_item, current_root[identifier_searched_item]

        if hasattr(current_root, identifier_searched_item):
            return identifier_searched_item, getattr(current_root, identifier_searched_item)

        raise ConfigurationNotRetrievable(f'Configuration option {identifier_searched_item} not found.')
