import copy
import datetime
import time

import msgpack

from functools import reduce
from typing import Any, Callable, Dict, Iterator, MutableSequence, Optional, Set, Tuple, Union, TypeVar, List, overload, \
    Iterable, Hashable, Sequence
from threading import Lock

from dataspree.inference.core.exceptions import ImplementationError, DataspreeInferenceException

Key = TypeVar('Key', bound=Hashable)
Val = TypeVar('Val')


class Item(Dict[Key, Val]):
    __uid: int = 0
    __uid_mutex: Lock = Lock()

    @staticmethod
    def default(id: Optional[str] = None, timestamp: Optional[int] = None, **kwargs: Any) -> \
            'Item[Union[str, Key], Union[Hashable, Val]]':
        """
        :return:  an item with default-initialized required fields if not specifically provided as arguments.
        """
        str_id: str
        if id is None:
            str_id = f'{Item.generate_uid()}'
        else:
            str_id = id

        if timestamp is None:
            timestamp = int(round(time.time() * 1000))

        return Item(id=str_id, timestamp=timestamp, **kwargs)

    @staticmethod
    def human_readable_time(timestamp_ms: Optional[int]) -> str:
        if timestamp_ms is None:
            timestamp_ms = int(time.time() * 1000)
        return datetime.datetime.fromtimestamp(timestamp_ms / 1000.0).isoformat()

    def human_readable_time_from_item_time(self: Dict[Key, Val]) -> str:
        return Item.human_readable_time(self['timestamp'])  # type: ignore

    _T = TypeVar('_T')

    @overload
    def get(self: Dict[Key, Val], key: Key, /) -> Optional[Val]:
        ...

    @overload
    def get(self: Dict[Key, Val], key: Key, default: Union[Val, _T], /) -> Union[Val, _T]:
        ...

    @overload
    def get(self: Dict[Key, Val], keys: List[Union[Key, Hashable]], /) -> Optional[Any]:
        ...

    @overload
    def get(self: Dict[Key, Val], keys: List[Union[Key, Hashable]], default: Union[Val, _T], /) -> Union[Any, Val, _T]:
        ...

    def get(self: Dict[Key, Val],
            keys: Union[Key, Key, List[Union[Key, Hashable]], List[Union[Key, Hashable]]],
            default: Any = None) -> Any:
        """
        If #keys holds a string, this function redirects to the dictionary's original `get`
        function. otherwise, redirect to #dict_get and return only the gathered value.

        Only dictionary keys are currently supported.

        :param keys: a single key (if str) or a list or iterator of keys.
        :param keys: single key or List or Iterator of keys which constitute the path in the #self
                     to the acquired.
        :param default: default value that is returned when the entry is not found.
        :return: self[keys[0]][keys[1]]...[keys[len(keys)]] or #default.
        """
        if not isinstance(keys, list):
            dict_key: Key = keys
            return dict.get(self, dict_key, default)

        found: bool
        value: Optional[Val]
        found, value = Item[Key, Val].find(self, keys)

        return value if found else default

    def set(self: Union[Dict[Key, Val]], keys: Union[Key, List[Hashable]], value: Any,
            update: Callable[[Val, Optional[Any]], Any] = lambda new, old: new) -> Any:
        """
        Update the #self's entry at #keys as indicated by the #value and the #update function:
        self[keys[0]]...[keys[len(keys)]] = update(value, self[keys[0]]...[keys[len(keys)]]).

        If some entry on the path does not exist yet, insert a dictionary.
        Only generally applicable if the entire path in the item consists of dictionaries.

        :param keys: single key or List of keys which constitute the path to the value to be set
                     in the #self.
        :param value: new value considered in the update function to then be registered in the
                      dictionary #self.
        :param update: update function, which allows to update the current entry .
                       By default, this function discards the old value and only applies the new
                       value.
        :return: the updated value
        """
        if not isinstance(keys, list):
            keys = [keys]
        if len(keys) == 0:
            raise ValueError('Item.set: Keys must not be empty.')
        # add the key if it does not exist and continue the traversal according to the keys.
        item_part = reduce(lambda b, key: Item.uni_setdefault(b, key, dict()), keys[:-1], self)
        # update the last part of the keys
        item_part[keys[-1]] = update(value, Item.uni_get(item_part, keys[-1], None))
        return item_part[keys[-1]]

    def map(self: Union[Dict[Key, Val]], keys: Union[MutableSequence[Any], Any], function: Callable[[Any], Any]) -> Any:
        """
        Update the #self's entry at #keys as indicated by the update #function:
        self[keys[0]]...[keys[len(keys)]] = update(self[keys[0]]...[keys[len(keys)]]).

        If some entry on the path does not exist yet, insert a dictionary.
        Only generally applicable if the entire path in the item consists of dictionaries.

        :param keys: single key or List of keys which constitute the path to the value to be set
                     in the #self.
        :param function: update function, which allows to update the current entry.
        :return: the updated value
        """
        return Item.set(self, keys, value=None, update=lambda _new, old: function(old))

    def find(self: Dict[Key, Val], keys: Sequence[Union[Key, Hashable]]) -> Tuple[bool, Optional[Any]]:
        """
        Return the value associated to the nested key of self[keys[0]][keys[1]]...[keys[len(keys)]]
        if found, and otherwise the default value.

        :param keys: single key or List or Iterator of keys which constitute the path in the #self
                     to the acquired.
        :return: self[keys[0]][keys[1]]...[keys[len(keys)]] or #d.
        """
        if not isinstance(keys, list) and not isinstance(keys, Iterator):
            raise ValueError('Item.find: wrong function call for key arguments')

        leaf = object()
        o = reduce(lambda el, key: leaf if el is leaf else Item.uni_get(el, key, leaf), keys, self)

        if o is not leaf:
            return True, o

        return False, None

    @overload
    def setdefault(self: Dict[Key, Val], key: Key, value: None = ..., /) -> Optional[Val]:
        ...

    @overload
    def setdefault(self: Dict[Key, Val], key: Key, value: Val, /) -> Val:
        ...

    @overload
    def setdefault(self: Dict[Key, Val], key: List[Union[Key, Hashable]], value: Any = None, /) -> Optional[Val]:
        ...

    def setdefault(self: Dict[Key, Val], keys: Union[Key, List[Union[Key, Hashable]]], value: Any = None, /) -> Any:
        """
        Set the #self's entry at #keys to #value if it does not exist yet:
                          self[keys[0]]...[keys[len(keys)]] = value
        If some entry on the path does not exist yet, insert a dictionary.

        :param keys: single key or List of keys which constitute the path to the value to be set
                     in the #self.
        :param value: new value to be set in case the old value does not exist or is None.
        @see Item.set()
        """
        if not isinstance(keys, list):
            return dict.setdefault(self, keys, value)  # type: ignore

        return Item[Key, Val].set(self, keys, value, lambda new, old: new if old is None else old)

    def set_add(self: Union[Dict[Key, Val]], keys: Union[MutableSequence[Any], Any], value) -> Any:
        """
        Accumulate the #self's entry at #keys or initialize it to #value if it does not exist yet.
        self[keys[0]]...[keys[len(keys)]] = value
        If some entry on the path does not exist yet, insert a dictionary.

        :param keys: single key or List of keys which constitute the path to the value to be set
                     in the #self.
        :param value: new value to be set in case the old value does not exist or is None.
        @see Item.set()
        """
        return Item.set(self, keys, value, lambda new, old: new if old is None else old + new)

    def transform(self: Union[Dict[Key, Val]], keys: Union[MutableSequence[Any], Any], update) -> Any:
        """
        Transform the element located at at the path indicated by #keys by applying the update
        function

        :param keys: single key or List of keys which constitute the path to the value to be set
                     in the #self.
        :param update: Function which takes the old value as a single argument and returns the
                       updated value. If the entry did not exist yet, the value passed to the
                       update function is None
        :return: The updated Value
        @see Item.set
        """
        Item.set(self, keys, None, lambda new, old: update(old))

    def drop(self: Dict[Key, Val], path: List[Key]):
        """
        Remove the element indicated by #path from the dictionary structure.

        If the element to be removed is an entry inside a list, set the value that the
        list assumes at that position to None.
        If it is an entry inside a dictionary, remove the key-value pair.

        :param path: List of keys which constitute the path in the #self to the element
                     that is supposed to be removed.
                     Has to contain at least one element for the removal to be potentially
                     successful.
        :return: If the element was found and thus removed successfully.
        """
        if len(path) == 0:
            return None

        found: bool
        pop_parent: Optional[Val]
        found, pop_parent = Item[Key, Val].find(self, path[:-1])
        if not found:
            return None

        dummy = object()
        value = Item.uni_get(pop_parent, path[-1], dummy)
        if value is dummy:
            return None

        if isinstance(pop_parent, dict):
            pop_parent.pop(path[-1])
        elif isinstance(pop_parent, list):
            pop_parent[path[-1]] = None  # type: ignore
        else:
            raise DataspreeInferenceException(f'Unknown type for removing content: {type(pop_parent)}')
        return value

    def list_entries(self: Union[Dict[Key, Val]], path: Optional[Union[MutableSequence, Tuple]] = None,
                     path_segments: Optional[MutableSequence[Union[MutableSequence, Tuple, Iterator]]] = None) -> \
            MutableSequence[Any]:
        """
        Utility function for walk which discards the path information and yields only the entries
        in a list.
        @see Item.walk
        """
        return list(Item.walk_entries(self=self, path=path, path_segments=path_segments))

    def walk_entries(self: Union[Dict[Key, Val]], path: Optional[Union[MutableSequence, Tuple]] = None,
                     path_segments: Optional[MutableSequence[Union[MutableSequence, Tuple, Iterator]]] = None) -> \
            Iterator[Any]:
        """
        Utility function for walk which discards the path information and yields only the entries
        in an iterator.
        @see Item.walk
        """
        return (entry for entry, _ in Item.walk(self=self, path=path, path_segments=path_segments))

    def list(self: Union[Dict[Key, Val]], path: Optional[Union[MutableSequence, Tuple]] = None,
             path_segments: Optional[MutableSequence[Union[MutableSequence, Tuple, Iterator]]] = None) -> \
            MutableSequence[Tuple[Any, MutableSequence[Any]]]:
        """
        Utility function for walk which transforms the iterator into a list.
        @see Item.walk
        """
        return list(Item.walk(self=self, path=path, path_segments=path_segments))

    def walk(self: Union[Dict[Key, Val]], path: Optional[Sequence[Hashable]] = None,
             path_segments: Optional[Sequence[Sequence[Hashable]]] = None) -> Iterator[Any]:
        """
        Iterates through the entries in #self over wildcard axes.

        Example:
        self.walk(['inference', 'detection', 'results', self.get_wildcard(), 'confidence'])

        Yields all the content of all confidence entries contained in any child of
        /inference/detection/results.

        As it might be important to be aware of the complete path for processing, (e.g. for
        applying a filter or group by operator afterwards or by identifying the nature of the
        return value, this function returns a tuple which contains both the content and the
        path specification (i.e., the replacements of the wildcards that yield the current self).

        :param path: Path which points to the elements to be iterated over.
                     Dimensions along which the self is to walk are to be indicated with a wildcard.
        :param path_segments: List of connected paths, separated by the start of a new path.
        :return: iterator[Tuple[self, path replacements]]
        @see Item.get_wildcard()
        """
        if (path is None) == (path_segments is None):
            raise ValueError('Path xor path_segment have to be specified.')

        path_segments_not_none: MutableSequence[Union[MutableSequence[Any], Tuple, Iterator]]
        if path_segments is None:
            path_segments_not_none = Item.list_split(path, Item.get_wildcard())
        else:
            path_segments_not_none = path_segments

        yield from Item.__walk(self, path_segments_not_none, [])

    @staticmethod
    def __walk(remaining_item: Union[Dict[Key, Val]], path: MutableSequence[Union[MutableSequence, Iterator, Tuple]],
               path_replacements: MutableSequence[Any]):

        if len(path) == 0:
            return

        # Acquire the dictionary entry on the current level.
        entry_found, branch = Item.find(remaining_item, path[0])

        if entry_found and len(path) > 0:
            if len(path) == 1:
                # Replaced all wildcards from the path, and yield the current value
                yield branch, path_replacements
            else:
                # Wildcard: Iterate over the current level
                items = branch.items() if type(branch) is dict else enumerate(branch)
                yield from (next_level for key, new_item in items for next_level in
                            Item.__walk(new_item, path[1:], path_replacements + [key]))

    def walk_leaves(self, current_path: Optional[MutableSequence] = None,
                    leaf_filter: Optional[Callable[[Any, Any], bool]] = None):
        return self.walk_leaves_and_branches(current_path, leaf_filter)

    def walk_leaves_and_branches(self, current_path: Optional[MutableSequence] = None,
                                 leaf_filter: Optional[Callable[[Any, Any], bool]] = None,
                                 branch_filter: Optional[Callable[[Any, Any], bool]] = None):
        """
        Iterate over all the leaves of #self.
        @return: key, value pair which represent the path to the current leaf and the value
                 it assumes.
        """

        if current_path is None:
            current_path = []

        # is subclass(type(self), abc.Iterable) would also apply for strings
        if issubclass(type(self), list) or issubclass(type(self), dict):
            items = self.items() if issubclass(type(self), dict) else enumerate(self)
            if branch_filter is not None and branch_filter(current_path, self):
                yield current_path, self
            else:
                yield from (next_level for key, new_item in items for next_level in
                            Item.walk_leaves_and_branches(new_item, current_path + [key], leaf_filter=leaf_filter,
                                                          branch_filter=branch_filter))
        else:

            if leaf_filter is None or leaf_filter(current_path, self):
                yield current_path, self

    @staticmethod
    def is_path_maintained(path: MutableSequence[str],
                           included_paths: Optional[List[List[Hashable]]] = None,
                           excluded_paths: Optional[List[List[Hashable]]] = None):
        """
        Returns True, if the #path is not filtered out from an item in the current configuration
        of included and excluded paths.

        @param included_paths: see Item.extract.
        @param excluded_paths: see Item.extract.
        @param path:  Path which is either filtered out or maintained when calling Item.extract.

        @see extract
        """
        item = Item()
        item.set(path, object())

        found, _ = Item.find(item, path)
        if not found:
            raise ImplementationError('Item not found after setting it.')

        found, _ = Item.find(item.extract(included_paths=included_paths, excluded_paths=excluded_paths,
                                          deep_copy=False, soft_copy=False)[1], path)

        return found

    def extract(self, included_paths: Optional[Sequence[Sequence[Hashable]]] = None,
                excluded_paths: Optional[Sequence[Sequence[Hashable]]] = None,
                deep_copy: bool = False, soft_copy=False, pack: bool = False) -> Tuple[bool, Any]:
        """
        Extract some elements out of #self and return them.

        @param included_paths: Paths to be manually included or None, which is a shortcut
                               for wildcard.

        @param excluded_paths: Paths to be manually stripped from resulting item prior to returning.

        @param pack:           If pack is true, attempt packing the item via msgpack.
                               If packing fails, return false in the first return argument

        @param deep_copy:      If deep_copy is true and pack is false or packing fails,
                               perform a deep copy of the returned elements.

        @param soft_copy:      If soft_copy is true and deep_copy and pack are false,
                               soft-copy the paths that are manually specified  as #included_paths.


        @return: Tuple: Packing succeeded (if enabled), extracted content from the current item.
        """
        soft_copy &= not deep_copy and not pack
        deep_copy &= not pack

        output: Item = Item()
        if included_paths is None:
            output = Item(self)

        if included_paths:
            for path in included_paths:
                for curr_element, curr_replacement in self.walk(path):
                    keys = list(Item.iter_replace(path, Item.get_wildcard(),
                                                  iter(curr_replacement)))

                    if soft_copy and type(curr_element) in [list, dict, Item]:
                        curr_element = type(curr_element)(curr_element)

                    output.set(keys, curr_element)

        pop_paths = list()
        if excluded_paths:
            for path in excluded_paths:
                for curr_element, curr_replacement in output.walk(path):
                    keys = list(Item.iter_replace(path, Item.get_wildcard(), iter(curr_replacement)))
                    pop_paths.append(keys)

        # Determine the paths that we need to copy in order to not interfere with the item.
        # Is the order relevant?
        copy_set: Set[Tuple] = set()
        for path in pop_paths:
            for i in range(1, len(path)):
                if not Item.find(output, path)[0]:
                    break

                path_tuple = tuple(path[:i])
                if path_tuple not in copy_set:
                    copy_set.add(path_tuple)
                    output.map(list(path_tuple), copy.copy)

        for pop_path in pop_paths:
            output.drop(pop_path)

        if pack:
            try:
                return True, msgpack.packb(output)

            except Exception:
                deep_copy = True

        if deep_copy:
            return False, copy.deepcopy(output)

        return True, output

    @staticmethod
    def list_split(arr: Iterable[Hashable], sep: Hashable) -> List[List[Hashable]]:
        """
        Split #arr on #sep and return a list of all sub-lists.
        The behavior is consistent with str.split

        In particular, when called with an empty list, this function therefore returns [[]], as
        there is only a single sub-list [], returned in the list of all sub-lists.
        Compare: ''.split('1') -> ['']

        In case the original list #arr contains only a separator element, the resulting list
        contains two empty lists. Compare: '1'.split('1') -> ['', '']

        :param arr: original iterator or list
        :param sep: entry which marks the beginning of a new sub-list
        :return:    List[List]]
        """
        return reduce(lambda a, b: a + [[]] if b == sep else a[:-1] + [a[-1] + [b]], arr, [[]])

    @staticmethod
    def iter_replace(arr: Sequence[str], separator: str, replacements: Iterator) -> Iterable[str]:
        """
        Replace all occurrences of the #separator in #arr by values from #replacements and return
        the result in a new iterator.
        Comparison is performed through the __eq__ function.

        :param arr: original iterator or list
        :param separator: entry which is to be replaced if found
        :param replacements: Iterator which yields at least one replacement for each separator in
                             #arr.
        :return: iterator which holds the updated list.
        """
        return map(lambda a: next(replacements) if a == separator else a, arr)

    @staticmethod
    def uni_get(structure, key, default=None):
        """
        Utility which serves as a unified interface to both lists and dictionaries to acquire data
        with a default fallback.
        Returns the fallback if the type is neither list-like or an implementation of dict.

        @param structure: the structure from which we would like to acquire data.
        @param key:       the key which yields the required data
        @param default:   default value which is returned if acquiring the data failed.
        @return:          the gathered value or the default value if acquiring the data failed.
        """
        if issubclass(type(structure), dict):
            return dict.get(structure, key, default)
        elif callable(getattr(structure, "__len__", None)) \
                and callable(getattr(structure, "__getitem__", None)):
            if type(key) is int and 0 <= key < len(structure):
                return structure[key]
        return getattr(structure, key, default)

    @staticmethod
    def uni_setdefault(structure, key, value):
        if isinstance(structure, dict):
            return structure.setdefault(key, value)

        elif callable(getattr(structure, "__len__", None)) \
                and callable(getattr(structure, "__getitem__", None)):

            if type(key) is int and 0 <= key < len(structure):
                return structure[key]

            raise ValueError(f'Cannot setdefault in list-like object of length {len(structure)} at postion {key}.')

        raise ValueError(f'Cannot set default in type {type(structure).__name__}.')

    @staticmethod
    def generate_uid() -> int:
        with Item.__uid_mutex:
            Item.__uid += 1
            return Item.__uid

    @staticmethod
    def get_wildcard() -> str:
        """
        :return: The wildcard that can be used for the walk functionality
        @see walk, get_entries
        """
        return '*'

    @staticmethod
    def get_item_uid_key():
        """
        :return: The key for placing the unique item identifier
        """
        return 'uuid_item'
