import logging
from abc import abstractmethod
from typing import Any, Callable, Dict, List, Optional, Union, Type

from dataspree.inference.core.exceptions import ImplementationError
from dataspree.inference.core.consumer import Consumer
from dataspree.inference.core.decorators import action
from dataspree.inference.core.item import Item
from dataspree.inference.core.producer import Producer
from dataspree.inference.core.config.configurable import Configurable

from dataspree.inference.core.protection import LicenseProtection


class Node(Configurable):
    __application: Optional['Application'] = None

    @staticmethod
    def register(application: 'Application'):
        Node.__application = application

    def __init__(self, name: str) -> None:
        Configurable.__init__(self)
        self.name: str = name

        self.logger = logging.getLogger(__name__)
        self.consumers: Dict[str, Consumer] = dict()
        self.producers: Dict[str, Producer] = dict()
        self.running: bool = False
        self.actions: Dict[str, Callable] = dict()

        # Gather all actions, getter, setter and updater
        for name in dir(self.__class__):
            func = getattr(self.__class__, name)

            if isinstance(func, action):
                self.actions[name] = func

        LicenseProtection.raise_node_instantiation_disallowed(self.get_config_key())

    def __update_producers(self) -> None:
        if self.__application is not None:
            self.__application.producers.update(self.producers)
            self.__application.consumers.update(self.consumers)
            self.__application.set_update_available()

    def add_consumer(self, consumer: Consumer) -> None:
        self.consumers[consumer.name] = consumer

    def add_producer(self, producer: Producer) -> None:
        if producer.name in self.producers:
            raise ImplementationError(f'Adding producer under the slot {producer.name} that is already registered.')
        self.producers[producer.name] = producer
        self.__update_producers()

    def remove_producer(self, producer: Producer) -> None:
        try:
            del self.producers[producer.name]

        except KeyError:
            raise ImplementationError(f'Attempting to remove a producer {producer.name} that does not exist.')

        if self.__application is not None:
            try:
                del self.__application.producers[producer.name]
            except KeyError:
                self.logger.warning(f'Attempting to remove a producer {producer.name} from application, where it was '
                                    f'not registered.')
        self.__update_producers()

    def update_producer(self, producer: Producer) -> None:
        self.producers[producer.name] = producer
        self.__update_producers()

    def start(self) -> None:
        try:
            self.running = True
            self.run()
        except SystemExit:
            self.logger.error(f'Node {self.name} attempted to stop inference. Intercepted and ignored.')

        finally:
            self.running = False

    def perform_action(self, action_name: str, args: List, kwargs: Dict) -> None:
        try:
            func: Optional[Callable] = getattr(self, action_name)
            if func is not None:
                func(self, *args, **kwargs)
        except AttributeError:
            pass

    @classmethod
    def get_config_entries(cls) -> Dict[str, Any]:
        """Return the properties with which this instances of this Node can be configured."""
        return dict()

    @staticmethod
    @abstractmethod
    def get_config_categories() -> List[str]:
        """Return the categories of the configuration of this kind of nodes."""

    @classmethod
    def get_config_key(cls) -> str:
        """
        Node implementations can implement this function by providing the configuration key
        that identifies their class in the configuration file.
        """
        return super(Node, cls).get_config_key()

    @classmethod
    def build(cls, config: Union[Dict, Item], name: Optional[str], from_user_input: bool = False) \
            -> Optional['Node']:
        """Create an instance of the node from a configuration."""
        return cls.default_build(config=config, name=name, from_user_input=from_user_input)  #

    @abstractmethod
    def run(self) -> None:
        """Run the processing loop."""

    def stop(self) -> None:
        """Stop the node."""
        self.running = False

    def disconnect(self) -> None:
        """Remove all connections to other nodes (producers and consumers)."""
        for consumer in self.consumers.values():
            consumer.disconnect_all_producers()
        for producer in self.producers.values():
            producer.disconnect_all_consumers()
