# mypy: ignore-errors
# -*- coding: utf-8 -*-
"""
Implementation of the Streaming API channel.

Author: GBC09 / BU05 / SW
SICK AG, Waldkirch
email: techsupport0905@sick.de

Copyright note: Redistribution and use in source, with or without modification, are permitted.

Liability clause: THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import logging
import select
import socket
import struct
import sys
import time

logger = logging.getLogger(__name__)


def to_hex(bStr):
    """ just to produce a readable output of the device responses """
    if not isinstance(bStr, bytes):
        raise RuntimeError("invalid protocol string (not a bytes object)")
    fStr = '==> hexDump\n'
    cnt = 0
    for b in bytearray(bStr):
        if cnt == 0:
            fStr += '    '
        cnt += 1
        fStr += '{:02X} '.format(b)
        if cnt % 4 == 0:
            fStr += ' '
        if cnt % 16 == 0:
            fStr += '\n'
            cnt = 0
    if cnt != 0:
        fStr += '\n'
    fStr += 'hexDump <=='
    return fStr


class PartialFrame:

    def __init__(self):
        self.filled_size: int = 0
        self.buffer: bytes = bytes()

    def reserve(self, n_bytes: int):
        additional_required_space: int = n_bytes - len(self.buffer)
        if additional_required_space < 0:
            raise RuntimeError('Partial frame corrupt.')

        if additional_required_space > 0:
            if self.filled_size != 0:
                raise RuntimeError('Partial frame corrupt: filled size != 0.')

            header_len: int = len(self.buffer)
            self.buffer = bytearray(self.buffer + bytearray(additional_required_space))
            self.filled_size = header_len

    def receive_into(self, sock, total_bytes: int):
        self.reserve(total_bytes)

        view = memoryview(self.buffer)
        while self.filled_size < total_bytes:
            bytes_received = sock.recv_into(view[self.filled_size:], total_bytes - self.filled_size)
            self.filled_size += bytes_received

            if bytes_received == 0:
                # premature end of connection
                raise RuntimeError(f'Received {self.filled_size} but requested {total_bytes} bytes')

    def read(self, sock, total_bytes: int):
        while len(self.buffer) < total_bytes:
            new_bytes = sock.recv(total_bytes - len(self.buffer))
            self.buffer += new_bytes

            if not len(new_bytes):
                raise socket.error(f'Network connection closed by peer. Requested {total_bytes} bytes, '
                                   f'received {len(self.buffer)} bytes.')

        return bytes(self.buffer[:total_bytes])

class Streaming:

    """ All methods that use the streaming channel. """
    def __init__(self, ipAddress='192.168.1.10', tcpPort=2114):
        self.ipAddress = ipAddress
        self.tcpPort = tcpPort
        self.sock_stream = None
        self.frame = None
        self.partial_frame = None


    ''' Opens the streaming channel. '''

    def openStream(self):
        logger.info("Opening streaming socket...")
        self.sock_stream = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock_stream.settimeout(5)
        try:
            self.sock_stream.connect((self.ipAddress, self.tcpPort))
        except socket.error as err:
            logger.error("Error on connecting to %s:%d: %s" % (self.ipAddress, self.tcpPort, err))
            sys.exit(2)
        logger.info("...done.")

    def closeStream(self):
        """ Closes the streaming channel. """
        if self.sock_stream is not None:
            logging.info("Closing streaming connection...")
            self.sock_stream.close()
            logging.info("...done.")

    def sendBlobRequest(self):
        """ Sending a blob request. """
        MSG_BLREQ_TX = b'BlbReq'

        logger.debug("Sending BlbReq: %s" % (to_hex(MSG_BLREQ_TX)))
        self.sock_stream.send(MSG_BLREQ_TX)

    def fileno(self):
        """ Make our stream compatible with select"""
        return self.sock_stream.fileno()

    @staticmethod
    def poll(stream_seq, timeout=None):
        """ Polls whether a frame is available on a sequence of streams

        Returns a tuple of
          1. a list of received frames; at the index of a passed stream is the received frame
          2. a list of streams in exceptional state
        or None if there was none"""

        rlist, _, _ = select.select(stream_seq, [], [], timeout)

        frames = []
        for stream in stream_seq:
            if stream in rlist:
                stream.getFrame(peek=True)
                frames.append(stream.frame)
            else:
                frames.append(None)
        return frames

    def get_most_frequent_frame(self, max_skipped_frames, *args, **kwargs) -> int:
        try:
            self.sock_stream.setblocking(False)
            last_frame = None
            for num_skipped_frames in range(max_skipped_frames):
                try:
                    self.getFrame(*args, **kwargs)
                    last_frame = self.frame

                except (BlockingIOError, socket.timeout) as e:
                    self.frame = last_frame
                    return num_skipped_frames

            return max_skipped_frames
        finally:
            self.sock_stream.setblocking(True)

    def getFrame(self, peek=False):
        """ Receives the raw data frame from the device via the streaming channel.

         peek(bool): if True it returns True if no data were found.
        """
        logger.debug('Reading image from stream...')
        self.frame = None  # reset old frame
        self.frame_acq_time_s = None
        if self.partial_frame is None:
            self.partial_frame = PartialFrame()

        keepRunning = True

        BLOB_HEAD_LEN = 11
        try:
            # may raise socket.error
            header = self.partial_frame.read(self.sock_stream, BLOB_HEAD_LEN)
            assert len(header) == BLOB_HEAD_LEN
        except (BlockingIOError, socket.timeout):
            header = None

        if not header:
            if peek:
                return
            raise socket.timeout(f'BLOB header received a timeout. '
                                 f'Loaded {len(self.partial_frame.buffer)} / {BLOB_HEAD_LEN}')

        self.frame_acq_time_s = time.time()
        try:
            logger.debug("len(header) = %d dump: %s" % (len(header), to_hex(header)))
        except Exception as e:
            print('error.', header, type(header))
            raise e

        # check if the header content is as expected
        (magicword, pkgLength, protocolVersion, packetType) = \
            struct.unpack('>IIHB', header)
        if magicword != 0x02020202:
            logger.error("Unknown magic word: %0x" % (magicword))
            keepRunning = False
        if protocolVersion != 0x0001:
            logger.error("Unknown protocol version: %0x" % (protocolVersion))
            keepRunning = False
        if packetType != 0x62:
            logger.error("Unknown packet type: %0x" % (packetType))
            keepRunning = False

        if not keepRunning:
            raise RuntimeError('something is wrong with the buffer')

        # -3 for protocolVersion and packetType already received
        # +1 for checksum
        toread = pkgLength - 3 + 1
        logger.debug("pkgLength: %d" % (pkgLength))
        logger.debug("toread: %d" % (toread))

        self.partial_frame.receive_into(self.sock_stream, len(header) + toread)
        try:

            self.sock_stream.settimeout(0.01)
        except (socket.timeout, BlockingIOError) as e:
            raise socket.timeout(f'Received a timeout when trying to load data: {e}. '
                                 f'Loaded {self.partial_frame.filled_size} / {len(header) + toread}.')
        self.frame = self.partial_frame.buffer
        self.partial_frame = PartialFrame()

        frame_acq_stop = time.time()
        self.frame_revc_time_s = (frame_acq_stop - self.frame_acq_time_s)
        logger.info("Receiving took %0.1f ms" % ((self.frame_revc_time_s) * 1000))
        # full frame should be received now
        logger.debug("...done.")