from albumentations import CenterCrop
from typing import Optional


class RelativeCenterCrop(CenterCrop):
    """
    Crop center
    """

    def __init__(self, aspect_ratio: float, always_apply=False, p=1.0):
        """
        :param aspect_ratio:   width / height
        """
        super().__init__(width=1, height=1, always_apply=always_apply, p=p)

        if aspect_ratio <= 0:
            raise ValueError('Aspect ratio has to be > 0')

        self.aspect_ratio: float = aspect_ratio
        self.__last_input_height: int = 0
        self.__last_input_width: int = 0

    def update_params(self, params, **kwargs):
        params = super(RelativeCenterCrop, self).update_params(params, **kwargs)
        self.__last_input_height = params['rows']
        self.__last_input_width = params['cols']

        if self.__last_input_height >= self.__last_input_width:
            self.width = self.__last_input_width
            self.height = max(int(round(self.width / self.aspect_ratio)), 1)
        else:
            self.height = self.__last_input_height
            self.width = max(int(round(self.height * self.aspect_ratio)), 1)

        return super(RelativeCenterCrop, self).update_params(params, **kwargs)

    def __check_after_update(self, rows: Optional[int] = None, cols: Optional[int] = None, **_kwargs):
        if rows is None and cols is None:
            raise NotImplementedError('Rows and Columns have to be provided by a compatible version of Albumentations. '
                                 'Make sure you use the recommended version of Albumentations and report this '
                                 'incompatibility to a DataSpree developer.')
        else:
            if rows != self.__last_input_height:
                raise NotImplementedError(f'Rows {rows} does not match expected height {self.__last_input_height}. '
                                     f'The current version of Albumentations is incompatible with the '
                                     f'{type(self).__name__} augmentation.'
                                     'Make sure that you use the recommended version of Albumentations and report '
                                     'this incompatibility to a DataSpree developer.')
            if cols != self.__last_input_width:
                raise NotImplementedError(f'Rows {cols} does not match expected height {self.__last_input_width}'
                                     f'The current version of Albumentations is incompatible with the '
                                     f'{type(self).__name__} augmentation.'
                                     'Make sure that you use the recommended version of Albumentations and report '
                                     'this incompatibility to a DataSpree developer.')

    def apply(self, img, **params):
        self.__check_after_update(**params)
        return super(RelativeCenterCrop, self).apply(img, **params)

    def apply_to_bbox(self, bbox, **params):
        self.__check_after_update(**params)
        return super(RelativeCenterCrop, self).apply_to_bbox(bbox, **params)

    def apply_to_keypoint(self, keypoint, **params):
        self.__check_after_update(**params)
        return super(RelativeCenterCrop, self).apply_to_keypoint(keypoint, **params)

    def get_transform_init_args_names(self):
        return ('aspect_ratio',)

    @classmethod
    def get_class_fullname(cls):
        return cls.__name__
